from .chai_bot import ChaiBot
from .types import LatestMessage, Update, Message, MessageKind
from .t_room import TRoom
from .packaging import package, Metadata
from .deployment import upload_and_deploy, wait_for_deployment, get_bot_status, share_bot
from .notebook_utils import IS_NOTEBOOK, setup_notebook



import logging
import sys
from typing import List
from typing_extensions import final

from .logger import ColoredFormatter
from .types import Update, Message


class ChaiBot:
    """Base chatbot class.

    Chatbots must be made as a subclass of ChaiBot.
    The subclass must then implement the on_message() method, with the same signature.
    """
    # When starting a conversation with the bot, the app first sends a message with this string.
    # By detecting and responding to this string, the bot can conveniently provide a custom conversation opener.
    # This attribute is only for convenience; overriding this string does not change the string the bot will receive.
    FIRST_MESSAGE_STRING = "__first"

    @final
    def __init__(self, uid: str = "LOCAL_DEV"):
        self.uid = uid
        self.logger = logging.getLogger(self.__class__.__name__)
        self.setup_logger()

        self.setup()

    async def on_message(self, update: Update) -> str:
        """Event called when the bot receives a message.

        @param update:
        @return:
            String containing the reply.
        """
        raise NotImplementedError

    async def get_messages(self, conversation_id: str) -> List[Message]:
        """Fetches a list of messages for the specified conversation.

        This function is provided when using bots in a :class:TRoom and in deployment.
        Do not override or create an implementation.

        @param conversation_id:
        @return:
            List of messages in the conversation, sorted by increasing timestamp (i.e. oldest first).
        """
        raise NotImplemented

    def setup_logger(self):
        formatter = ColoredFormatter.default_chai_formatter()
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        self.logger.handlers = []
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.INFO)

    def setup(self):
        """Any setup should be done by overriding this function."""
        pass



import imghdr
import inspect
import json
import os
import pprint
import re
import shutil
import stat
import tempfile
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Type, AnyStr, List, Optional
from zipfile import ZipFile, ZIP_DEFLATED

import pkg_resources
import requests

from .chai_bot import ChaiBot

MAX_SUPPORTED_MEMORY = 4096


@dataclass
class Metadata:
    """Information required for bot deployment."""
    # Name of the bot.
    name: str
    # Profile image for the bot. Has to be a valid URL.
    image_url: str
    # The alphanumeric part of a hex color code. (E.g. ffffff)
    color: str
    # Developer Unique ID.
    developer_uid: str
    # Description of the bot.
    description: str
    # Python class (N.B. not object!) that inherits from ChaiBot.
    input_class: Type[ChaiBot]

    # Total available memory for the bot in MB. This includes memory needed to store sources and data.
    memory: int = 256

    def verify(self):
        """Performs basic checks to ensure validity of the metadata."""
        assert isinstance(self.name, str)
        assert len(self.name) >= 3, "Bot name has to be at least 3 characters"

        try:
            verify_image_url(self.image_url)
        except Exception:
            raise ValueError(f"Could not verify image url ({self.image_url})")

        assert isinstance(self.color, str)
        assert re.search(r"^(?:[0-9a-fA-F]{3}){1,2}$", self.color), \
            f"Color has to be provided as the alphanumeric part of the hex code (e.g. ffffff), found {self.color}"

        assert isinstance(self.memory, int), f"Attribute .memory has to be an integer (found type {type(self.memory)})."
        assert self.memory <= MAX_SUPPORTED_MEMORY, f"Attribute .memory has to be less than or equal to {MAX_SUPPORTED_MEMORY} (found {self.memory})."


def package(metadata: Metadata, requirements: Optional[List[str]] = None):
    """Packages the chatbot into a single archive for deployment.

    Performs some preliminary checks on the metadata.
    Creates a package.zip file in the directory containing the file that contains the bot class.

    :param metadata:
    :param requirements:
    :return:
    """
    print("Running verification checks on metadata.")
    metadata.verify()

    bot_file = Path(inspect.getfile(metadata.input_class))

    metadata_dict = {
        'name': metadata.name,
        'imageUrl': metadata.image_url,
        'color': metadata.color,
        'developerUid': metadata.developer_uid,
        'description': metadata.description,
        'inputFile': bot_file.stem,
        'inputClass': metadata.input_class.__name__,
        'memory': metadata.memory,
    }
    print("Prepared metadata:")
    pprint.pprint(metadata_dict)

    print("Preparing temporary directory...")
    with tempfile.TemporaryDirectory() as temp_dir:
        # Copy files in bot directory
        def ignore(src, names):
            # Do not store pycache folder.
            if '__pycache__' in names:
                return ['__pycache__']
            ignore_list = []
            if src == temp_dir:
                for name in names:
                    if name.startswith("_"):
                        warnings.warn(
                            f"Ignoring file at bot root directory with leading underscore in name: {name}.",
                            RuntimeWarning
                        )
                        ignore_list.append(name)
                    if name == "main.py":
                        raise RuntimeError("Bot root directory cannot contain a main.py file.")
            return ignore_list

        copytree(bot_file.parent, temp_dir, ignore=ignore)
        # Write metadata.json
        with (Path(temp_dir) / "metadata.json").open("w") as f:
            json.dump(metadata_dict, f)

        # Write requirements.txt
        if requirements:
            write_valid_requirements_file(Path(temp_dir) / "requirements.txt", requirements)

        # Create zip
        zip_path = bot_file.parent / "_package.zip"
        with zip_path.open("wb") as f:
            zipfile_from_folder(temp_dir, f)
        print(f"Created zip package at {zip_path}.")


def verify_image_url(url: str):
    """Verifies that the provided url resolves to an image.

    Performs a GET request on the given url and performs a trivial (non-conclusive) check
    that the image type can be inferred from the received bytes.

    :param url:
    :return:
    """
    r = requests.get(url)
    try:
        imghdr.what(None, h=r.content)
    except Exception:
        raise ValueError(
            f"Could not verify image type from bytes "
            f"(response content-type of {r.headers.get('content-type')})"
        )


def zipfile_from_folder(folder: AnyStr, file):
    # Adapted from https://stackoverflow.com/a/17080988
    with ZipFile(file, "w", compression=ZIP_DEFLATED) as zip_archive:
        for root, dirs, files in os.walk(folder):
            # add directory (needed for empty dirs)
            zip_archive.write(root, os.path.relpath(root, folder))
            for file in files:
                filename = os.path.join(root, file)
                if os.path.isfile(filename):  # regular files only
                    arcname = os.path.join(os.path.relpath(root, folder), file)
                    zip_archive.write(filename, arcname)


def copytree(src, dst, symlinks=False, ignore=None):
    """Copies files from src to dst.

    Taken from https://stackoverflow.com/a/22331852.
    Necessitated by Python 3.7 environment; Python 3.8's shutil.copytree can be used directly
    as it has the necessary dirs_exist_ok argument.

    :param src: Source directory
    :param dst: Target directory
    :param symlinks:
    :param ignore: Callable
    :return:
    """
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    if ignore:
        excl = ignore(src, lst)
        lst = [x for x in lst if x not in excl]
    for item in lst:
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if symlinks and os.path.islink(s):
            if os.path.lexists(d):
                os.remove(d)
            os.symlink(os.readlink(s), d)
            try:
                st = os.lstat(s)
                mode = stat.S_IMODE(st.st_mode)
                os.lchmod(d, mode)
            except Exception:
                pass  # lchmod not available
        elif os.path.isdir(s):
            copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def write_valid_requirements_file(path: AnyStr, requirements: List[str]):
    """Writes a valid requirements.txt file.

    Iterates through list of requirements, writing valid entries to the specified file,
    Ignores (and prints) invalid requirements.

    :param path:
    :param requirements:
    :return:
    """
    with Path(path).open("w") as f:
        for requirement in requirements:
            try:
                pkg_resources.Requirement.parse(requirement)
                f.write(requirement + "\n")
            except Exception as e:
                print(f"Ignoring requirement {requirement}: {e}")



import logging
import sys
from typing import Optional, Dict

from colorama import init, Fore, Back, Style

if sys.platform.startswith("win"):
    init()


class ColoredFormatter(logging.Formatter):
    """Colored log formatter.
    Adapted from https://gist.github.com/joshbode/58fac7ababc700f51e2a9ecdebe563ad.
    """

    def __init__(self, *args, colors: Optional[Dict[str, str]] = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record) -> str:
        """Format the specified record as text."""

        record.color = self.colors.get(record.levelname, '')
        record.reset = Style.RESET_ALL

        return super().format(record)

    @classmethod
    def default_chai_formatter(cls):
        return cls(
            '{asctime} |{color} {levelname:8} {reset}| {name} | {message}',
            style='{', datefmt='%Y-%m-%d %H:%M:%S',
            colors={
                'DEBUG': Fore.CYAN,
                'INFO': Fore.GREEN,
                'WARNING': Fore.YELLOW,
                'ERROR': Fore.RED,
                'CRITICAL': Fore.RED + Back.WHITE + Style.BRIGHT,
            }
        )



import time
from pathlib import Path
from typing import AnyStr, Optional

import requests
import segno
from halo import Halo
from typing_extensions import TypedDict

from .defaults import DEFAULT_BOT_STATUS_ENDPOINT, DEFAULT_SIGNED_URL_CREATOR, GUEST_KEY, GUEST_UID
from .notebook_utils import IS_NOTEBOOK, show_qr

PREVIOUS_DEPLOYMENTS = {}


def upload_and_deploy(package: AnyStr, uid: str = GUEST_UID, key: str = GUEST_KEY, bot_uid: str = None) -> str:
    """Uploads the given archive, triggering deployment of the chatbot.

    :param package: Path to the packaged chatbot zip.
    :param uid: Developer Unique Identifier. Defaults to a guest UID.
    :param key: Developer key. Defaults to a valid key for the guest UID.
    :param bot_uid: Used to modify an existing bot: the UID of the previously-deployed bot.
    :return bot_uid: The UID of the deployed bot
    """
    package = Path(package)

    if package in PREVIOUS_DEPLOYMENTS:
        previous_bot_uid = PREVIOUS_DEPLOYMENTS[package]
        print("Detected previous deployment from this location. Use the same bot UID as before?")
        print(f" [y] (default) Yes. Update the bot ({previous_bot_uid}).")
        print(f" [n] No. Deploy as a new bot.")
        input_key = input().lower()
        if input_key == "y" or input_key == "":
            bot_uid = previous_bot_uid
            print(f"Using previous bot UID: {bot_uid}")
        elif input_key == "n":
            pass
        else:
            raise RuntimeError("Unknown input.")

    try:
        r = requests.get(
            DEFAULT_SIGNED_URL_CREATOR,
            params={'uid': uid, 'key': key, 'bot_uid': bot_uid}
        )
        try:
            r.raise_for_status()
        except Exception:
            raise RuntimeError(r.json())
    except Exception:
        raise RuntimeError("Failed to retrieve signed URL.")

    url = r.text
    if bot_uid is None:
        print("Creating new bot.")
        bot_uid = parse_signed_url_for_bot_uid(url)
        print(f"Received bot UID: {bot_uid}")
    PREVIOUS_DEPLOYMENTS[package] = bot_uid

    with package.open("rb") as f:
        r = requests.put(url, data=f, headers={'content-type': 'application/zip'})
        r.raise_for_status()
    print(f"Successfully uploaded {package}.")
    return bot_uid


def parse_signed_url_for_bot_uid(url: str):
    """Parses a bot UID from a signed URL.

    Assumes the signed URL follows the following scheme:
        [...]/{bot_uid}.zip?[...]

    :param url:
    :return:
    """
    endpoint = url.split("?", maxsplit=1)[0]
    file_name = endpoint.split("/")[-1]
    bot_uid = file_name.split(".")[0]
    return bot_uid


SPINNER = None


def wait_for_deployment(bot_uid: str, sleep: float = 3):
    """Waits for deployment of the bot to complete.

    :param bot_uid:
    :param sleep: Polling interval in seconds.
    :return:
    """
    MIN_SLEEP = 1
    sleep = max(MIN_SLEEP, sleep)

    BOT_DEPLOYMENT_PROCESS = [
        "signed_url_created",
        "processing_upload",
        "deploying",
        "initialized",
    ]
    current_deployment_process = -1  # Index for BOT_DEPLOYMENT_PROCESS
    completed_processes = []
    existing_status = None
    try:
        existing_status = get_bot_status(bot_uid)
        print(f"Found previous deployment: Version {existing_status['activeDeployment']['version']}")
    except Exception:
        pass

    MAXIMUM_ERROR_RETRIES = 10
    error_retries = 0

    global SPINNER
    if SPINNER is not None:
        SPINNER.stop()
    SPINNER = Halo(text="Polling for progress...")
    SPINNER.start()

    time.sleep(5)  # Initial wait to avoid bot status error.

    while True:
        try:
            status = get_bot_status(bot_uid)
        except Exception as e:
            SPINNER.warn(f"Error getting bot status (uid: {bot_uid}): {e}")
            error_retries += 1
            if error_retries > MAXIMUM_ERROR_RETRIES:
                SPINNER.fail()
                print(f"Hit retry-on-error limit ({MAXIMUM_ERROR_RETRIES}).")
                break
            continue
        if existing_status is not None:
            # Check if new timestamp is later than existing timestamp
            if status['timestamp'] <= existing_status['timestamp']:
                # Do not parse old version
                continue
        status_str = status['status']
        if status_str not in BOT_DEPLOYMENT_PROCESS:
            raise ValueError(f"Unknown status: {status_str}.")
        new_current_deployment_process = BOT_DEPLOYMENT_PROCESS.index(status_str)
        if new_current_deployment_process != current_deployment_process:
            # Completed new step(s)
            for step in BOT_DEPLOYMENT_PROCESS[:new_current_deployment_process + 1]:
                if step in completed_processes:
                    continue
                SPINNER.succeed(step)
                completed_processes.append(step)
            current_deployment_process = new_current_deployment_process
            if current_deployment_process + 1 < len(BOT_DEPLOYMENT_PROCESS):
                # If next step exists, set spinner to next step
                SPINNER.start(
                    f"Waiting for next step: {BOT_DEPLOYMENT_PROCESS[current_deployment_process + 1]}"
                )
            else:
                # Next step does not exist; wait for final deployment confirmation
                SPINNER.start("Waiting for active_deployment confirmation...")
        if 'activeDeployment' in status:
            new_active_deployment = status['activeDeployment']
            if existing_status is None \
                    or ('activeDeployment' in existing_status
                        and new_active_deployment['version'] > existing_status['activeDeployment']['version']):
                SPINNER.succeed("active_deployment")
                print(f"New active deployment: {new_active_deployment}")
                break
        time.sleep(sleep)


class ActiveDeployment(TypedDict):
    timestamp: int
    version: int


class BotStatus(TypedDict):
    status: str
    timestamp: int
    activeDeployment: Optional[ActiveDeployment]


def get_bot_status(bot_uid: str) -> BotStatus:
    """Gets the status of the bot.

    :param bot_uid:
    :return:
    """
    r = requests.get(DEFAULT_BOT_STATUS_ENDPOINT, params={'bot_uid': bot_uid})
    r.raise_for_status()
    return r.json()


def share_bot(bot_uid: str) -> str:
    """Displays the url, a QR code, along with additional guidance.

    :param bot_uid:
    :return: The url for the bot.
    """
    url = f"chai://chai.ml/{bot_uid}"

    qr_code = segno.make_qr(url)
    print("Scan the QR code with your phone to start a chat in the app!")
    print(f"Or check it out at {url}")

    if IS_NOTEBOOK:
        show_qr(qr_code)
    else:
        qr_code.terminal()

    def save():
        path = input("Save QR code to: (Press [Enter] for default: 'qr.png') ")
        if len(path) == 0:
            path = "qr.png"
        qr_code.save(path, scale=10)
        print(f"Saved QR code to {path}.")

    def open_():
        qr_code.show(scale=10)

    actions = {
        "s": (save, "Save this QR code to an image file.")
    }
    if not IS_NOTEBOOK:
        actions["o"] = (open_, "Open QR code in external viewer.")

    while True:
        print("\nEnter one of the following keys to perform additional actions (or [Enter] to exit):")
        for key, action in actions.items():
            print(f" [{key}] {action[1]}")

        input_key = input().lower()
        if input_key in actions:
            actions[input_key][0]()
        else:
            print("Exiting.")
            break

    return url


advertise_deployed_bot = share_bot



import setuptools

with open("README.md", "r") as f:
    readme = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="chaipy",
    version="0.1.31",
    author="Chai",
    author_email="dev@chai.ml",
    description="A developer interface for creating chatbots for the Chai app.",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires='>=3.7',
    install_requires=requirements,
    extras_require={
        "notebook": ["nest_asyncio"],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Communications :: Chat',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: Apache Software License'
    ]
)
